# Hitachi air conditioner remote RAR-8P2

from ..remote import Remote

class Hitachi_RAR_8P2:
    format = {
      "format": {
        "preamble": [
          8,
          4
        ],
        "coding": "ppm",
        "zero": [
          1,
          1
        ],
        "one": [
          1,
          3
        ],
        "postamble": [
          1
        ],
        "pre_data": "01 10 00 40 BF FF 00 CC 33",
        "post_data": "03 FC 01 FE 88 77 00 FF 00 FF FF 00 FF 00 FF 00 FF 00",
        "byte_by_byte_complement": True,
        "timebase": 420,
        "gap": 49000
      },
      "keys": {}
    }

    power = False
    mode = 'blow'
    temperature = 24
    air_speed = 'auto'
    power_saving = False
    clean_inside = False

    def __init__(self, gpio):
        self.remote = Remote(self.format, gpio)

    def send(self):
        data = []
        if self.air_speed == 1:
            data.append(0x98)
        elif self.clean_inside:
            data.append(0xA3)
        elif self.air_speed == 5:
            data.append(0xA9)
        else:
            data.append(0x92)

        data.append(0x13)
        data.append(self.temperature * 4)
        data += [0] * 5

        if self.mode == 'blow':
            d = 1
        elif self.mode == 'cool':
            d = 3
        elif self.mode == 'dry':
            d = 5
        elif self.mode == 'heat':
            d = 6
        else:
            raise ValueError('mode must be one of blow, cool, dry or heat')

        if self.air_speed == 'auto':
            d |= 0x50
        elif self.air_speed <= 4:
            d |= self.air_speed * 0x10
        elif self.air_speed == 5:
            d |= 0x60
        else:
            raise ValueError('air_speed must be 1 to 5 or "auto"')
        data.append(d)

        if self.power:
            d = 0xf0
        else:
            d = 0xe0
        if not self.power_saving:
            d |= 1
        data.append(d)

        if self.air_speed == 5:
            data.append(3)
        else:
            data.append(0)

        data.append(0)

        if self.clean_inside:
            data.append(0x88)
        else:
            data.append(0x80)

        self.remote.send_data(bytes(data))



import json, os, sys
from argparse import ArgumentParser
from .io import receive
from .decode import decode
from .util import hexify
from .prettify import prettify
from .remote import Remote
from . import __version__

def bytes_to_hex(x):
    if isinstance(x, bytes):
        return hexify(x)
    raise TypeError

def do_receive(args):
    return receive(args.gpio, glitch=args.glitch, timeout=args.timeout)

def receive_and_decode(args):
    while True:
        data = decode(
            do_receive(args),
            tolerance = args.tolerance,
            min_gap = args.gap * 1000,
            min_pulses = args.pulses,
        )
        if data:
            return data

def record_key(args, name):
    print(f'Press the key named "{name}".')
    data1 = receive_and_decode(args)
    print('Press the same key again to verify.')
    data2 = receive_and_decode(args)
    if data2 == data1:
        return data2
    while True:
        print('Press the same key again to verify.')
        data3 = receive_and_decode(args)
        if data3 == data1 or data3 == data2:
            return data3
        data1 = data2
        data2 = data3

def record(args):
    keys = {}
    if os.path.exists(args.file):
        print(f'Loading "{args.file}"...')
        remote = Remote(args.file, None)
        keys = remote.unprettify()
    if args.keys:
        for name in args.keys:
            keys[name] = record_key(args, name)
    else:
        while True:
            name = input('Name of the key (blank to finish): ')
            if not name:
                break
            keys[name] = record_key(args, name)
    data = prettify(keys, carrier=int(round(args.carrier * 1000)))
    json.dump(data, open(args.file, 'w', encoding='utf8'), indent=2)
    print(f'Saved to "{args.file}".')

def play(args):
    remote = Remote(
        args.file,
        args.gpio,
        active_low = args.active_low,
        duty_cycle = args.duty_cycle,
    )
    for key in args.keys:
        remote.send(key, repeat=args.repeat)

def dump(args):
    keys = {}

    while True:
        if args.raw:
            pulses = do_receive(args)
            print(' '.join(str(x) for x in pulses))
        elif args.intermediate:
            data = receive_and_decode(args)
            print(json.dumps(data, default=bytes_to_hex))
        else:
            data = receive_and_decode(args)
            key = len(keys)
            keys[key] = data
            remote = Remote(prettify(keys), None)
            restored = remote.restore_data(remote.keys[key], False)
            if len(restored) == 1:
                restored = restored[0]
            print(json.dumps(restored, indent=2, default=bytes_to_hex))

def main():
    common_parser = ArgumentParser(add_help=False)
    common_parser.add_argument(
        '-g',
        '--gpio',
        help = 'GPIO to use',
        type = int,
        required = True,
    )

    decode_parser = ArgumentParser(add_help=False)
    decode_parser.add_argument(
        '-G',
        '--glitch',
        help = 'Ignore glitch shorter than GLITCH microseconds (default 100)',
        type = int,
        default = 100,
    )
    decode_parser.add_argument(
        '-t',
        '--timeout',
        help = 'Finish receiving after TIMEOUT milliseconds of silence (default 100)',
        type = int,
        default = 100,
    )
    decode_parser.add_argument(
        '-T',
        '--tolerance',
        help = 'Timing tolerance for decoding (default 0.2)',
        type = float,
        default = .2,
    )
    decode_parser.add_argument(
        '--gap',
        help = 'Minimum gap between data in milliseconds (default 15)',
        type = int,
        default = 15,
    )
    decode_parser.add_argument(
        '-p',
        '--pulses',
        help = 'Ignore signal with less than PULSES pulses (>=3, default 10)',
        type = int,
        default = 10,
    )

    file_parser = ArgumentParser(add_help=False)
    file_parser.add_argument(
        '-f',
        '--file',
        required = True,
        help = 'Filename to record/play',
    )

    root_parser = ArgumentParser(
        description = 'IR remote control for Raspberry Pi'
    )
    root_parser.set_defaults(func=None)
    subparsers = root_parser.add_subparsers(title='Sub commands')
    root_parser.add_argument(
        '-V',
        '--version',
        action='version',
        version=f'%(prog)s {__version__}',
    )


    record_parser = subparsers.add_parser(
        'record',
        help = 'Record IR to file',
        parents = [common_parser, file_parser, decode_parser],
    )
    record_parser.set_defaults(func=record)
    record_parser.add_argument(
        'keys',
        nargs = '*',
        metavar = 'KEY',
        help = 'Name of keys',
    )
    record_parser.add_argument(
        '-c',
        '--carrier',
        type = float,
        default = 38,
        help = 'Frequency of carrier wave in KHz (default 38). '
            'NOTE: This option does not affect receiving/decoding, '
            'but is recorded in the file.',
    )

    play_parser = subparsers.add_parser(
        'play',
        help = 'Play IR from file',
        parents = [common_parser, file_parser],
    )
    play_parser.set_defaults(func=play)
    play_parser.add_argument(
        'keys',
        nargs = '+',
        metavar = 'KEY',
        help = 'Name of keys',
    )
    play_parser.add_argument(
        '-r',
        '--repeat',
        metavar = 'N',
        type = int,
        help = 'Repeat sending each key N times (default 1)',
    )
    play_parser.add_argument(
        '-l',
        '--active-low',
        action = 'store_true',
        help = 'Drive LED as active low',
    )
    play_parser.add_argument(
        '-d',
        '--duty-cycle',
        type = float,
        default = 1/2,
        help = 'Duty cycle of carrier wave (default 0.5)',
    )

    dump_parser = subparsers.add_parser(
        'dump',
        help = 'Receive IR and dump',
        parents = [common_parser, decode_parser],
    )
    dump_parser.set_defaults(func=dump)
    dump_mode = dump_parser.add_mutually_exclusive_group()
    dump_mode.add_argument(
        '-r',
        '--raw',
        action = 'store_true',
        help = 'Print raw pulses',
    )
    dump_mode.add_argument(
        '-i',
        '--intermediate',
        action = 'store_true',
        help = 'Print intermediate data',
    )

    args = root_parser.parse_args()

    if args.func is None:
        root_parser.print_help()
        exit(1)

    if getattr(args, 'pulses', 10) < 3:
        print(f'PULSES must be >= 3', file=sys.stderr)
        exit(1)

    args.func(args)

if __name__ == '__main__':
    main()



import json, os
from time import time, sleep
from io import IOBase
from .encode import encode
from .io import send
from .util import bytes_to_bits, bits_to_bytes

class Remote:
    def __init__(self, data, gpio, active_low=False, duty_cycle=None):
        self.gpio = gpio
        self.active_low = active_low
        self.duty_cycle = duty_cycle
        self.last_sent = 0

        if isinstance(data, (str, os.PathLike)):
            data = open(data)

        if isinstance(data, IOBase):
            data = json.load(data)

        self.load(data)

    def load(self, data):
        self.formats = data.get('formats')
        if not self.formats:
            self.formats = [data['format']]
        self.keys = data['keys']
        self.cache = {}

    def restore_data(self, data, restore_data=True):
        if not isinstance(data, list):
            data = [data]
        result = []
        for part in data:
            if isinstance(part, dict):
                d = part['data']
                format = part['format']
            else:
                d = part
                format = 0

            if isinstance(d, str):
                d = bytes.fromhex(d)

            r = self.formats[format].copy()

            if restore_data and r['coding'] != 'raw':
                pre = r.get('pre_data')
                if pre: r['pre_data'] = bytes.fromhex(pre)

                post = r.get('post_data')
                if post: r['post_data'] = bytes.fromhex(post)

                if r.get('byte_by_byte_complement'):
                    d = bytes(sum(zip(d, (byte ^ 0xff for byte in d)), ()))

            r['data'] = d

            result.append(r)
        return result

    def encode(self, data):
        data = self.restore_data(data)
        pulses = encode(data)
        gap = data[-1].get('gap')
        carrier = data[-1].get('carrier')
        return pulses, gap, carrier

    def send_pulses(self, pulses, gap, carrier, repeat):
        t = (gap or 50_000) / 1e6 - (time() - self.last_sent)
        if t > 0:
            sleep(t)
        self.last_sent = time()
        send(
            self.gpio,
            pulses,
            active_low = self.active_low,
            duty_cycle = self.duty_cycle,
            carrier = carrier,
            repeat = repeat,
            gap = gap,
        )

    def send_data(self, data, repeat=1):
        pulses, gap, carrier = self.encode(data)
        self.send_pulses(pulses, gap, carrier, repeat)

    def send(self, key, repeat=1):
        data = self.keys[key]
        cache = self.cache.get(key)
        if cache:
            pulses, gap, carrier = cache
        else:
            pulses, gap, carrier = self.encode(data)
            self.cache[key] = pulses, gap, carrier
        self.send_pulses(pulses, gap, carrier, repeat)

    def unprettify(self):
        result = {}
        for name, data in self.keys.items():
            parts = self.restore_data(data)
            r_parts = []
            for part in parts:
                r = part.copy()
                if part['coding'] == 'raw':
                    r['data'] = (
                        part.get('pre_data', []) +
                        part['data'] +
                        part.get('post_data', [])
                    )
                else:
                    bits = []

                    pre = part.get('pre_data')
                    if pre:
                        bits += bytes_to_bits(
                            pre,
                            part,
                            part.get('pre_data_bits'),
                        )

                    bits += bytes_to_bits(part['data'], part)

                    post = part.get('post_data')
                    if post:
                        bits += bytes_to_bits(
                            post,
                            part,
                            part.get('post_data_bits'),
                        )

                    r['data'] = bits_to_bytes(bits, part.get('msb_first'))
                    r['bits'] = (
                        part.get('pre_data_bits', 0) +
                        part.get('bits', len(part['data']) * 8) +
                        part.get('post_data_bits', 0)
                    )
                r.pop('pre_data', None)
                r.pop('pre_data_bits', None)
                r.pop('post_data', None)
                r.pop('post_data_bits', None)
                r.pop('byte_by_byte_complement', None)
                r_parts.append(r)
            result[name] = r_parts
        return result



from threading import Lock
from functools import wraps

def reverse_bits(byte):
    return int(f'{byte:08b}'[::-1], 2)

def bits_to_bytes(bits, msb_first):
    data = []
    b = 0
    byte = 0
    for bit in bits:
        if bit:
            byte |= 1 << b
        b += 1
        if b >= 8:
            data.append(byte)
            byte = 0
            b = 0
    if b:
        data.append(byte)
    if msb_first:
        data = [reverse_bits(byte) for byte in data]
    return bytes(data)

def bytes_to_bits(data, format, bits=None):
    bits = bits or format.get('bits') or (len(data) * 8)
    msb_first = format.get('msb_first')
    result = []
    for byte in data:
        if msb_first:
            byte = reverse_bits(byte)
        for bit in range(8):
            result.append(byte & 1)
            if len(result) >= bits:
                return result
            byte >>= 1
    return result

def hexify(data):
    return ' '.join(f'{byte:02X}' for byte in data)

def synchronized(fn):
    lock = Lock()
    @wraps(fn)
    def wrapper(*args, **kws):
        with lock:
            return fn(*args, **kws)
    return wrapper



from .util import bytes_to_bits, bits_to_bytes, hexify

def most_frequent(xs):
    d = {}
    for x in xs:
        d.setdefault(x, 0)
        d[x] += 1
    max = 0
    result = None
    for value, count in d.items():
        if count > max:
            max = count
            result = value
    return result

def find_pre_post(format, data):
    bitwise = format.get('bits', 0) % 8 != 0

    if bitwise:
        data = [bytes_to_bits(d, format) for d in data]

    pre = data[0]
    post = data[0]
    for d in data[1:]:
        for i in range(len(pre)):
            if pre[i] != d[i]:
                pre = pre[:i]
                break
        for i in range(1, len(post)+1):
            if post[-i] != d[-i]:
                if i > 1:
                    post = post[-i+1:]
                else:
                    post = []
                break

    if len(pre) == len(data[0]): return
    if len(post) == len(data[0]): return

    if bitwise:
        if pre:
            format['pre_data_bits'] = len(pre)
            format['bits'] -= len(pre)
            pre = bits_to_bytes(pre, format.get('msb_first'))
        if post:
            format['post_data_bits'] = len(post)
            format['bits'] -= len(post)
            post = bits_to_bytes(post, format.get('msb_first'))

    if pre: format['pre_data'] = pre
    if post: format['post_data'] = post

def remove_pre_post(keys, formats):
    for name, data in keys.items():
        for part in data:
            f = part['format']
            format = formats[f]

            pre_bits = post_bits = 0
            pre = format.get('pre_data', [])
            if pre:
                pre_bits = format.get('pre_data_bits', 0)

            post = format.get('post_data', [])
            if post:
                post_bits = format.get('post_data_bits', 0)

            if not pre and not post:
                continue

            if pre_bits or post_bits: # bitwise?
                length = pre_bits + post_bits + format['bits']
                bits = bytes_to_bits(part['data'], format, length)
                if pre_bits:
                    bits = bits[pre_bits:]
                if post_bits:
                    bits = bits[:-post_bits]
                part['data'] = bits_to_bytes(bits, format.get('msb_first'))
            else: # bytewise
                data = part['data']
                data = data[len(pre):]
                if post:
                    data = data[:-len(post)]
                part['data'] = data

            if format.get('byte_by_byte_complement'):
                part['data'] = part['data'][::2]

def remove_complement(keys, formats):
    complement = [True] * len(formats)
    for name, data in keys.items():
        for part in data:
            f = part['format']
            if not complement[f]:
                continue

            format = formats[f]

            if format.get('bits', 0) % 8:
                complement[f] = False
                continue

            d = part['data']
            even = d[::2]
            odd = d[1::2]
            for e, o in zip(even, odd):
                if e ^ o != 0xff:
                    complement[f] = False
                    break

    for c, format in zip(complement, formats):
        if c:
            format['byte_by_byte_complement'] = True

    for name, data in keys.items():
        for part in data:
            f = part['format']
            if not complement[f]:
                continue
            part['data'] = part['data'][::2]

def prettify_keys(keys, formats):
    result = {}
    for name, data in keys.items():
        parts = []
        is_raw = True
        for part in data:
            if formats[part['format']]['coding'] != 'raw':
                is_raw = False
            d = part['data']
            if isinstance(d, bytes):
                d = hexify(d)
            part['data'] = d
            if part['format'] == 0:
                part = part['data']
            parts.append(part)
        if len(parts) == 1 and not is_raw:
            parts = parts[0]
        result[name] = parts
    return result

def prettify(raw_keys, carrier=38_000):
    formats = []
    timebases = []
    gaps = []
    keys = {}
    data_per_format = []
    for name, raw_data in raw_keys.items():
        data_list = []
        for d in raw_data:
            format = d.copy()
            data = format.pop('data')
            timebase = format.pop('timebase')
            gap = format.pop('gap')

            for k, v in list(format.items()):
                if not v:
                    del format[k]

            if format.get('bits', 1) % 8 == 0:
                del format['bits']

            for i, f in enumerate(formats):
                if f == format:
                    f_index = i
                    break
            else:
                formats.append(format)
                f_index = len(formats) - 1
                timebases.append([])
                gaps.append([])
                data_per_format.append([])

            timebases[f_index].append(timebase)
            if gap:
                gaps[f_index].append(gap)

            data_list.append(dict(format=f_index, data=data))

            data_per_format[f_index].append(data)
        keys[name] = data_list

    for i, format in enumerate(formats):
        find_pre_post(format, data_per_format[i])
    remove_pre_post(keys, formats)

    remove_complement(keys, formats)

    keys = prettify_keys(keys, formats)

    for i, format in enumerate(formats):
        format['timebase'] = most_frequent(timebases[i])
        if gaps[i]:
            format['gap'] = most_frequent(gaps[i])

        if format['coding'] != 'raw':
            pre = format.get('pre_data')
            if pre:
                format['pre_data'] = hexify(pre)

            post = format.get('post_data')
            if post:
                format['post_data'] = hexify(post)

        format['carrier'] = carrier

    if len(formats) == 1:
        return dict(format=formats[0], keys=keys)
    else:
        return dict(formats=formats, keys=keys)



from time import sleep, time
import pigpio
from .util import synchronized

def receive(gpio, glitch=100, timeout=200):
    last_tick = 0
    in_code = False
    recording = False
    pulses = []

    def callback(gpio, level, tick):
        nonlocal last_tick, in_code, recording
        usec = pigpio.tickDiff(last_tick, tick)
        last_tick = tick
        if level == pigpio.TIMEOUT:
            if in_code:
                in_code = False
                recording = False
        else:
            if in_code:
                pulses.append(usec)
            else:
                # Ignore the first one (time since last timeout).
                in_code = True

    pi = pigpio.pi() # Connect to Pi.

    if not pi.connected:
       raise IOError

    pi.set_mode(gpio, pigpio.INPUT) # IR RX connected to this GPIO.
    pi.set_glitch_filter(gpio, glitch) # Ignore glitches.
    pi.set_watchdog(gpio, timeout)
    pi.callback(gpio, pigpio.EITHER_EDGE, callback)

    while True:
        recording = True
        while recording:
            sleep(0.01)
        if pulses:
            break

    pi.stop()

    return pulses

@synchronized
def send(
    gpio,
    pulses,
    active_low = False,
    carrier = None,
    duty_cycle = None,
    repeat = None,
    gap = None,
):
    carrier = carrier or 38_000
    duty_cycle = duty_cycle or 1/2
    repeat = repeat or 1
    gap = gap or 50_000

    def generate_carrier(length):
        wf = []
        cycle = 1_000_000 / carrier
        cycles = int(round(length / cycle))
        on = int(round(cycle * duty_cycle))
        sofar = 0
        for c in range(cycles):
            target = int(round((c+1)*cycle))
            sofar += on
            off = target - sofar
            sofar += off
            if active_low:
                wf.append(pigpio.pulse(0, 1<<gpio, on))
                wf.append(pigpio.pulse(1<<gpio, 0, off))
            else:
                wf.append(pigpio.pulse(1<<gpio, 0, on))
                wf.append(pigpio.pulse(0, 1<<gpio, off))
        return wf

    pi = pigpio.pi() # Connect to Pi.
    if not pi.connected:
        raise IOError
    pi.set_mode(gpio, pigpio.OUTPUT)

    pi.wave_clear()
    wave_ids = {}
    chain = []

    if len(pulses) & 1:
        pulses = pulses + [0]

    if repeat > 1:
        pulses[-1] += gap

    for i in range(0, len(pulses), 2):
        mark = pulses[i]
        space = pulses[i + 1]
        wave_id = wave_ids.get((mark, space))
        if not wave_id:
            wf = generate_carrier(mark)
            if space:
                wf.append(pigpio.pulse(0, 0, space))
            pi.wave_add_generic(wf)
            wave_id = pi.wave_create()
            wave_ids[(mark, space)] = wave_id
        chain.append(wave_id)

    for i in range(repeat):
        pi.wave_chain(chain)
        while pi.wave_tx_busy():
            sleep(0.001)

    pi.wave_clear()
    pi.stop()



import math
from .util import bits_to_bytes

def round_to_2_digits(x):
    if not x: return x
    if x < 0: return -round_to_2_digits(abs(x))
    frac = 1 - math.floor(math.log10(x))
    y = round(x, frac)
    if frac <= 0:
        return int(y)
    return y

def split_pulses(pulses, min_gap, min_pulses):
    parts = []
    part = []
    gaps = []

    for i, pulse in enumerate(pulses):
        if (i & 1) and pulse >= min_gap:
            gaps.append(pulse)
            if len(part) >= min_pulses:
                parts.append(part)
            part = []
        else:
            part.append(pulse)
    if len(part) >= min_pulses:
        parts.append(part)

    if gaps:
        gap = sorted(gaps)[len(gaps) // 2]
    else:
        gap = None

    return parts, gap

def min_pulse(pulses, tolerance):
    groups = []
    max_pulse = None
    n = 0
    total = 0
    sorted_pulses = sorted(pulses)
    max_pulse = sorted_pulses[0] * (1 + 2 * tolerance)
    for pulse in sorted_pulses:
        if pulse > max_pulse:
            if n >= 3:
                break
            else:
                n = 0
                total = 0
                max_pulse = pulse * (1 + 2 * tolerance)
        n += 1
        total += pulse
    if not n:
        return 0
    return total / n

def try_to_quantize(parts, tolerance, timebase, deviation, strict):
    result = []
    for part in parts:
        result_part = []
        for i, pulse in enumerate(part):
            if i & 1:
                pulse += deviation
            else:
                pulse -= deviation
            f = pulse / timebase
            n = int(round(f))
            if n <= 0:
                return None
            e = abs(f - n)
            if strict:
                if e > tolerance:
                    return None
            else:
                if e > tolerance * n:
                    return None
            result_part.append(n)
        result.append(result_part)
    return result

def quantize(parts, tolerance):
    marks = []
    spaces = []
    for part in parts:
        marks += part[0::2]
        spaces += part[1::2]
    min_mark = min_pulse(marks, tolerance)
    min_space = min_pulse(spaces, tolerance)

    for strict in (True, False):
        min_dev = float('infinity')
        best = None
        for m_len, s_len in ((1, 1), (1, 2), (2, 1)):
            timebase = (min_mark + min_space) / (m_len + s_len)
            deviation = (
                (min_mark - timebase * m_len) +
                (timebase * s_len - min_space)
            ) / 2
            result = try_to_quantize(
                parts,
                tolerance,
                timebase,
                deviation,
                strict,
            )
            if not result: continue
            if abs(deviation) < min_dev:
                min_dev = abs(deviation)
                best = timebase, result
        if best:
            return best

    # Failed to quantize
    return 1, parts

def remove_repeats(parts):
    i = len(parts) - 1
    while i > 0:
        if parts[i] == parts[i - 1]:
            del parts[i]
        i -= 1

# Pulse Position Modulation
def decode_ppm(pulses, msb_first=False):
    preamble = []
    for pre, pulse in enumerate(pulses):
        if pulse == 1:
            break
        preamble.append(pulse)
    if len(preamble) >= len(pulses):
        return None

    codes = set()
    for i in range(pre, len(pulses) - 1, 2):
        pair = pulses[i], pulses[i+1]
        codes.add(pair)
        if len(codes) >= 2:
            break
    if len(codes) != 2: return None

    zero, one = sorted(codes)
    bits = []
    byte_separator = None
    for i in range(pre, len(pulses) - 1, 2):
        pair = pulses[i], pulses[i+1]
        if pair == one:
            bits.append(1)
        elif pair == zero:
            bits.append(0)
        elif pair != zero:
            if len(bits) % 8 == 0:
                byte_separator = pair
                continue
            break
    else:
        i += 2

    if len(bits) < 5:
        return None

    postamble = pulses[i:]
    return dict(
        preamble = preamble,
        coding = 'ppm',
        zero = list(zero),
        one = list(one),
        byte_separator = byte_separator,
        msb_first = msb_first,
        bits = len(bits),
        data = bits_to_bytes(bits, msb_first),
        postamble = postamble,
    )

# Manchester coding
def decode_manchester(
    pulses,
    skip,
    preamble,
    msb_first = True,
    zero = (0, 1),
    one = (1, 0),
    long_bit = None,
):
    levels = []
    level = 0
    for pulse in pulses:
        level = 1 - level
        levels += [level] * pulse
    levels = levels[skip:]
    if len(levels) & 1:
        levels.append(0)

    if long_bit is not None:
        if len(levels) <= long_bit * 2 + 1:
            return None
        del levels[(long_bit + 1) * 2]
        del levels[long_bit * 2]

    bits = []
    for i in range(0, len(levels) - 1, 2):
        pair = levels[i], levels[i+1]
        if pair == one:
            bits.append(1)
        elif pair == zero:
            bits.append(0)
        else:
            return None

    return dict(
        preamble = preamble,
        coding = 'manchester',
        zero = list(zero),
        one = list(one),
        long_bit = long_bit,
        msb_first = msb_first,
        bits = len(bits),
        data = bits_to_bytes(bits, msb_first),
    )

def decode_raw(pulses):
    return dict(coding='raw', data=pulses)

def decode(
    pulses,
    min_gap = 15_000,
    tolerance = 0.2,
    min_pulses = 10,
):
    if len(pulses) < min_pulses:
        return None

    if (len(pulses) & 1) == 0:
        return None

    parts, gap = split_pulses(pulses, min_gap, min_pulses)
    if not parts:
        return None

    timebase, parts = quantize(parts, tolerance)
    if timebase == 1:
        return None
    remove_repeats(parts)

    results = []
    for part in parts:
        result = decode_ppm(part)
        if not result:
            if part[:3] == ([6, 2, 1]):
                # RC6
                result = decode_manchester(part, 10, [6, 2, 1, 1], long_bit=3)
            elif part[:3] == [1, 5, 1]:
                # NOKIA
                result = decode_manchester(
                    part,
                    8,
                    [1, 5, 1, 1],
                    msb_first = False,
                )
            else:
                # RC5
                result = decode_manchester(
                    part,
                    1,
                    [1],
                    zero = (1, 0),
                    one = (0, 1)
                )
        if not result:
            result = decode_raw(part)
        results.append(result)

    timebase = round_to_2_digits(timebase)
    gap = round_to_2_digits(gap)

    for result in results:
        result['timebase'] = timebase
        result['gap'] = gap

    return results



from setuptools import setup
from pathlib import Path

readme = open(Path(__file__).parent / 'README.md', encoding='utf8').read()

setup(
    name = 'PiIR',
    description = 'IR remote control for Raspberry Pi',
    long_description = readme,
    long_description_content_type = 'text/markdown',
    author = 'Takeshi Sone',
    author_email = 'takeshi.sone@gmail.com',
    install_requires = ['pigpio'],
    python_requires = '>= 3.6',
    url = 'https://github.com/ts1/PiIR',
    license = 'MIT',
    packages = ['piir', 'piir.remotes'],
    entry_points = {
        'console_scripts': ['piir=piir.cli:main'],
    },
    tests_require = ['pytest'],
    setup_requires = ['pytest-runner'],
    classifiers = [
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Topic :: System :: Hardware',
        'Topic :: Home Automation',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3 :: Only',
    ],
)



from .util import bytes_to_bits

def data_to_binary(code):
    result = []

    pre = code.get('pre_data')
    if pre:
        result += bytes_to_bits(pre, code, code.get('pre_data_bits'))

    result +=  bytes_to_bits(code['data'], code, code.get('bits'))

    post = code.get('post_data')
    if post:
        result += bytes_to_bits(post, code, code.get('post_data_bits'))

    return result

def encode_ppm(code):
    pulses = []
    pulses += code.get('preamble', [])
    data = data_to_binary(code)
    zero = code['zero']
    one = code['one']
    bits = 0
    byte_separator = code.get('byte_separator')
    for bit in data:
        if bit:
            pulses += one
        else:
            pulses += zero
        bits += 1
        if byte_separator and (bits % 8) == 0 and bits < len(data):
            pulses += byte_separator
    pulses += code.get('postamble', [])
    return pulses

def encode_manchester(code):
    levels = []
    level = 1
    for pulse in code.get('preamble', []):
        levels += [level] * pulse
        level = 1 - level

    data = data_to_binary(code)
    zero = code['zero']
    one = code['one']
    long_bit = code.get('long_bit')
    for i, bit in enumerate(data):
        if i == long_bit:
            if bit:
                levels += [one[0], one[0], one[1], one[1]]
            else:
                levels += [zero[0] + zero[0] + zero[1] + zero[1]]
        else:
            if bit:
                levels += one
            else:
                levels += zero

    pulses = []
    level = 1
    pulse = 0
    for l in levels:
        if l != level:
            level = l
            pulses.append(pulse)
            pulse = 0
        pulse += 1
    if pulse:
        pulses.append(pulse)

    return pulses

def encode_raw(code):
    return code['data']

encoders = dict(
    ppm = encode_ppm,
    manchester = encode_manchester,
    raw = encode_raw,
)

def encode_single(code):
    coding = code['coding']
    encoder = encoders.get(coding)
    if not encoder:
        raise ValueError(f'Unknown coding "{coding}"')
    pulses = encoder(code)

    t = code['timebase']
    pulses = [p * t for p in pulses]

    gap = code.get('gap') or 50_000
    if len(pulses) & 1:
        pulses.append(gap)
    else:
        pulses[-1] += gap

    return pulses

def encode(codes):
    pulses = []
    for code in codes:
        pulses += encode_single(code)
    if (len(pulses) & 1) == 0:
        del pulses[-1]
    return pulses
